class Extension:
    """Metadata for an extension.

    This describes the configuration options for an extension, and it provides
    a way to configure the extension.
    """

    def __init__(self, name, description, activate, config_options=()):
        self._name = name
        self._description = description
        self._config_options = tuple(config_options)
        self._activate = activate

    @property
    def name(self):
        "The name of the extension."
        return self._name

    @property
    def description(self):
        "A description of the extension."
        return self._description

    def activate(self, full_config, extension_config):
        return self._activate(full_config, extension_config)

    @property
    def config_options(self):
        "An iterable of Options for the extension."
        return self._config_options

    def __repr__(self):
        return "Extension(name='{}')".format(self.name)



"""Define a simple system of extension-points and extensions.

An extension-point represents a point in a program that can be extended. It comprises a name and a collection of
extensions at that point. Each extension on an extension-point provides a specific "implementation" for that
extension-point.

Extensions themselves comprise a name, a description, a sequence of configurable options, and a method for "activating"
them. This activation produces the object that actually embodies the implementation of the extension.
"""

import logging

from .extension import Extension
from .option import Option

log = logging.getLogger()


class ExtensionPoint:
    """A named point in a program that can be extended.
    """

    def __init__(self, name):
        self._name = name
        self._extensions = {}

    @property
    def name(self):
        "The name of the extension point."
        return self._name

    def add(self, name, description, activate, config_options=()):
        """Add an extension to the point.
        """
        extension = Extension(
            name=name,
            description=description,
            config_options=config_options,
            activate=activate)

        if extension.name in self._extensions:
            raise ValueError(
                'Extension {} already in {}'.format(extension.name, self))

        self._extensions[extension.name] = extension

    def activate(self, name, config):
        """Activate an extension.

        Args:
            name: The name of the extension to activate.
            config: The full config dict.

        Returns:
            The activate extension object (i.e. as returned from the extension on activation).

        Raises:
            KeyError: There is no extension named `name`.
        """
        return self._extensions[name].activate(
            config,
            config.get(self.name, {}).get(name, {}))

    def config_options(self):
        """All configurable options for this extension point.

        Iterable of Options.
        """
        for extension in self._extensions.values():
            path = (self.name, extension.name)
            for option in extension.config_options:
                yield Option(
                    name=option.name,
                    description=option.description,
                    default=option.default,
                    # Each option path is "translated" into the overall config
                    path=option.path + path
                )

    def __getitem__(self, name):
        """Get the extension object with the given name.

        Returns: An Extension instance.

        Raises:
            KeyError: There is no extension with that name.
        """
        return self._extensions[name]

    def __iter__(self):
        "Iterable of extension names."
        return iter(self._extensions.values())

    def __repr__(self):
        return "ExtensionPoint(name='{}')".format(self.name)



class Profile:
    """A collection of configuration options and extension points.
    
    This is essentially an "application configuration", combining configuration details along
    with the available extensions (inside extension points).
    """
    def __init__(self):
        self._options = set()
        self._extension_points = _ExtensionPoints()

    def add_option(self, option):
        self._options.add(option)

    def options(self):
        yield from self._options
        for ep in self.extension_points:
            yield from ep.config_options()

    @property
    def extension_points(self):
        return self._extension_points


class _ExtensionPoints:
    def __init__(self):
        self._eps = {}

    def add(self, extension_point):
        self._eps[extension_point.name] = extension_point

    def __getitem__(self, name):
        return self._eps[name]

    def __iter__(self):
        return iter(self._eps.values())



import logging

import pkg_resources
from stevedore import ExtensionManager

from .extension_point import ExtensionPoint

log = logging.getLogger()


def load_from_pkg_resources(namespace):
    """Discover ExtensionPoints using pkg_resources.

    This uses pkg_resources to scan the entry-points for `namespace`. Each dotted entrypoint name should have two parts.
    The first will be `namespace` and the second will be assumed to be the name of the entry point to which it belongs.

    Each discovered entry point should be a callable that takes a single argument of an ExtensionPoint. The extension
    should populate the ExtensionPoint appropriately, e.g. by calling `ExtensionPoint.add()`.

    Args:
        namespace: The pkg_resources namespace to scan.

    Returns: An iterable of ExtensionPoints.
    """
    for entry_point_name in pkg_resources.get_entry_map(namespace):
        toks = entry_point_name.split('.')
        if len(toks) == 1:
            # This happens for e.g. console_scripts
            continue

        # TODO: Do we need to extend for the case of an entry_point_name with more than two tokens? We could return a
        # tree of ExtensionPoint instead of just a flat mapping.

        assert toks[0] == namespace

        extension_point = ExtensionPoint(toks[1])

        manager = ExtensionManager(
            '.'.join(toks),
            on_load_failure_callback=_log_extension_loading_failure)

        for extension in manager:
            extension.plugin(extension_point)

        yield extension_point


def _log_extension_loading_failure(_mgr, extension_point, err):
    # We have to log at the `error` level here as opposed to, say, `info`
    # because logging isn't configure when we reach here. We need this infor to
    # print with the default logging settings.
    log.error('Plugin load failure: extension-point="%s", err="%s"',
              extension_point, err)
